/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.entities.base;

import com.jogamp.newt.event.InputEvent;
import com.jogamp.opengl.GL3bc;
import fr.amap.lidar.amapvox.engine3d.util.InputStatus;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * Abstract class for "behaviour" classes that can 
 * be added to a scene to perform some actions.
 * @author Tristan Muller (tristan.muller@cirad.fr)
 */
@ThreadSafe
public abstract class Behaviour {
    
    /**
     * If this has been initialized.
     */
    @GuardedBy("this")
    private volatile boolean initialized;

    /**
     * Create a behaviour with initialized set to false.
     */
    public Behaviour() {
        this.initialized = false;
    }
    
    /**
     * Initialize this behaviour by calling the onInit method.
     * @param gl The OpenGL context object.
     */
    public final synchronized void init(final GL3bc gl) {
        if (!initialized) {
            initialized = true;
            onInit(gl);
        }
    }
    
    /**
     * If that has been initialized, perform inputting actions of this 
     * behaviour by calling the onUpdate method.
     * @param event The input event.
     * @param status The input status.
     */
    public final synchronized void inputting(final InputEvent event, final InputStatus status) {
        if (initialized) {
            onInputting(event,status);
        }
    }
    
    /**
     * If that has been initialized, update this behaviour by 
     * calling the onUpdate method ; or call the init method before.
     * @param gl The OpenGL context object.
     * @param delta The time elapsed between the last frame and this one.
     */
    public final synchronized void update(final GL3bc gl, final float delta) {
        if (!initialized) {
            init(gl);
        }
        
        onUpdate(gl, delta);
    }
    
    /**
     * If that has been initialized, display this behaviour by 
     * calling the onDisplay method ; or call the init method before.
     * @param gl The OpenGL context object.
     */
    public final synchronized void display(final GL3bc gl) {
        if (!initialized) {
            init(gl);
        }
        
        onDisplay(gl);
    }
    
    /**
     * If that has been initialized, rescale this behaviour by 
     * calling the onRescale method ; or call the init method before. 
     * @param gl The OpenGL context object.
     * @param width The new sceen weight.
     * @param height The new screen height.
     */
    public final synchronized void rescale(final GL3bc gl, final int width, final int height) {
        if (!initialized) {
            init(gl);
        }
        
        onRescale(gl, width, height);
    }
    
    /**
     * Destroy this behaviour.
     * @param gl The OpenGL context object.
     */
    public final synchronized void destroy(final GL3bc gl) {
        if (initialized) {
            initialized = false;
            onDestroy(gl);
        }
    }
    
    /**
     * Implement it to specify what appends during the initialization.
     * @param gl The OpenGL context object.
     */
    public abstract void onInit(final GL3bc gl);
    
    /**
     * Implement it to specify what appends when an input event is received.
     * @param event The NEWT input event.
     * @param status The input status type (release, click ...)
     */
    public abstract void onInputting(final InputEvent event, final InputStatus status);
    
    /**
     * Implement it to specify what appends when this behaviour is updated.
     * @param gl The OpenGL context object.
     * @param delta The time elapsed between the last frame and this one.
     */
    public abstract void onUpdate(final GL3bc gl, final float delta);
    
    /**
     * Implement it to specify what appends when this behaviour is displayed.
     * @param gl The OpenGL context object.
     */
    public abstract void onDisplay(final GL3bc gl);
    
    /**
     * Implement it to specify what appends when this behaviour is rescaled.
     * @param gl The OpenGL context object.
     * @param width The new sceen weight.
     * @param height The new sceen height.
     */
    public abstract void onRescale(final GL3bc gl, final int width, final int height);
    
    /**
     * Implement it to specify what appends when this behaviour is destroy.
     * @param gl The OpenGL context object.
     */
    public abstract void onDestroy(final GL3bc gl);
    
}
