/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.resources.materials;

import org.joml.Matrix4f;
import org.joml.Quaternionf;
import org.joml.Vector3f;
import fr.amap.lidar.amapvox.engine3d.resources.programs.Program;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * Class to store data for a shader that is linked 
 * with a renderer entity and drawn by a program.
 */
@ThreadSafe
public abstract class Material {
        
    /**
     * The translation vector.
     */
    @GuardedBy("this")
    private Vector3f position;
    
    /**
     * The rotation quaternion.
     */
    @GuardedBy("this")
    private Quaternionf rotation;
    
    /**
     * The scale vector.
     */
    @GuardedBy("this")
    private Vector3f scale;
    
    /**
     * The camera transformation matrix.
     */
    @GuardedBy("this")
    private Matrix4f view;
    
    /**
     * The camera projection matrix.
     */
    @GuardedBy("this")
    private Matrix4f projection;
    
    /**
     * The camera position.
     */
    @GuardedBy("this")
    private Vector3f viewPosition;
    
    /**
     * If this material is set has hiden.
     */
    @GuardedBy("this")
    private boolean hiden;
        
    /**
     * Create a material.
     */
    public Material() {
        this.position = new Vector3f();
        this.rotation = new Quaternionf();
        this.view = new Matrix4f().identity();
        this.projection = new Matrix4f().identity();
        this.viewPosition = new Vector3f();
        this.scale = new Vector3f(1, 1, 1);
    }
    
    /**
     * Generate an appropriate program of this material.
     * @return 
     */
    public abstract Program generateProgram();

    /**
     * Change the position vector.
     * @param position 
     */
    public final synchronized void setPosition(final Vector3f position) {
        this.position = position;
    }

    /**
     * Change the rotation vector.
     * @param rotation 
     */
    public final synchronized void setRotation(final Quaternionf rotation) {
        this.rotation = rotation;
    }

    /**
     * Change the current camera view matrix.
     * @param view 
     */
    public final synchronized void setView(final Matrix4f view) {
        this.view = view;
    }

    /**
     * Change the current camera projection matrix.
     * @param projection 
     */
    public final synchronized void setProjection(final Matrix4f projection) {
        this.projection = projection;
    }

    /**
     * Set the scale value.
     * @param scale 
     */
    public final synchronized void setScale(final Vector3f scale) {
        this.scale = scale;
    }

    /**
     * Set if this material is hiden.
     * @param hiden 
     */
    public final synchronized void setHiden(final boolean hiden) {
        this.hiden = hiden;
    }
    /**
     * Get the position vector.
     * @return 
     */
    public final synchronized Vector3f getPosition() {
        return new Vector3f(position);
    }

    /**
     * Get the rotation quaternion.
     * @return 
     */
    public final synchronized Quaternionf getRotation() {
        return new Quaternionf(rotation);
    }

    /**
     * Get the current camera view matrix.
     * @return 
     */
    public final synchronized Matrix4f getView() {
        return new Matrix4f(view);
    }

    /**
     * Get the current camera projection matrix.
     * @return 
     */
    public final synchronized Matrix4f getProjection() {
        return new Matrix4f(projection);
    }
    
    /**
     * Get the current camera position.
     * @return 
     */
    public final synchronized Vector3f getViewPosition() {
        return new Vector3f(viewPosition);
    }

    /**
     * Get the scale value.
     * @return 
     */
    public final synchronized Vector3f getScale() {
        return scale;
    }

    /**
     * Get if the visibility is set to hiden.
     * @return 
     */
    public final synchronized boolean isHiden() {
        return hiden;
    }

    /**
     * Chnage the current camera position.
     * @param viewPosition 
     */
    public final synchronized void setViewPosition(final Vector3f viewPosition) {
        this.viewPosition = viewPosition;
    }

}
