/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.resources.meshes;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import net.jcip.annotations.ThreadSafe;

/**
 * Class to store geometry data.
 */
@ThreadSafe
public final class Geometry {

    private final CopyOnWriteArrayList<Float> vertices;
    private final CopyOnWriteArrayList<Float> normals;
    private final CopyOnWriteArrayList<Float> texcoords;
    private final CopyOnWriteArrayList<Integer> vertexIndices;
    private final CopyOnWriteArrayList<Integer> normalIndices;
    private final CopyOnWriteArrayList<Integer> texcoordIndices;

    /**
     * Create an empty geometry object.
     */
    public Geometry() {
        this.vertices = new CopyOnWriteArrayList();
        this.normals = new CopyOnWriteArrayList();
        this.texcoords = new CopyOnWriteArrayList();
        this.vertexIndices = new CopyOnWriteArrayList();
        this.normalIndices = new CopyOnWriteArrayList();
        this.texcoordIndices = new CopyOnWriteArrayList();
    }

    /**
     * Load an OBJ file and parse it as a Geometry object.
     * @param resourcePath
     * @return
     * @throws IOException 
     */
    public static final Geometry loadFromOBJ(final String resourcePath) throws IOException {
        final Geometry geometry = new Geometry();
        
        final InputStream in = Geometry.class.getResourceAsStream(resourcePath);
        final InputStreamReader isr = new InputStreamReader(in);
        final BufferedReader reader = new BufferedReader(isr);

        String line;
        
        while ((line = reader.readLine()) != null) {
            if (line.startsWith("v ")) {
                final String[] split = line.substring(2).split(" ");
                geometry.addVertex(
                        Float.valueOf(split[0]),
                        Float.valueOf(split[1]),
                        Float.valueOf(split[2])
                );
            } else if (line.startsWith("vt")) {
                final String[] split = line.substring(3).split(" ");
                geometry.addTexcoord(
                        Float.valueOf(split[0]),
                        Float.valueOf(split[1])
                );
            } else if (line.startsWith("vn")) {
                final String[] split = line.substring(3).split(" ");
                geometry.addNormal(
                        Float.valueOf(split[0]),
                        Float.valueOf(split[1]),
                        Float.valueOf(split[2])
                );
            } else if (line.startsWith("f")) {
                final String[] split = line.substring(2).split(" ");
                
                for (int i = 0; i < split.length; i++) {
                    final String[] indices = split[i].split("/");
                    geometry.addIndex(
                            Integer.valueOf(indices[0]) - 1, 
                            Integer.valueOf(indices[1]) - 1,
                            Integer.valueOf(indices[2]) - 1
                    );
                }
            }
        }
        
        return geometry;
    }

    /**
     * Add a vertex to the current vertices list.
     * @param x
     * @param y
     * @param z 
     */
    public final void addVertex(final float x, final float y, final float z) {
        vertices.add(x);
        vertices.add(y);
        vertices.add(z);
    }
    
    /**
     * Add a normal to the current list.
     * @param x
     * @param y
     * @param z 
     */
    public final void addNormal(final float x, final float y, final float z) {
        normals.add(x);
        normals.add(y);
        normals.add(z);
    }
    
    /**
     * Add a texcoord to the current list.
     * @param x
     * @param y 
     */
    public final void addTexcoord(final float x, final float y) {
        texcoords.add(x);
        texcoords.add(y);
    }
    
    /**
     * Add an index for each data type (vertex, texcoord and normal)
     * @param vertex
     * @param texcoord
     * @param normal 
     */
    public final void addIndex(final int vertex, final int texcoord, final int normal) {
        vertexIndices.add(vertex);
        texcoordIndices.add(texcoord);
        normalIndices.add(normal);
    }

    /**
     * Get a copy of the vertices in an array.
     * @return 
     */
    public final float[] getVertices() {
        /* Create an unmodifiable copy of the list to make it thread safe */
        final List<Float> unmodifiable = Collections.unmodifiableList(vertices);
        final List<Integer> unmodifiableIndices = Collections.unmodifiableList(vertexIndices);
        final float[] values = new float[unmodifiableIndices.size() * 3];
        
        for (int i = 0; i < unmodifiableIndices.size(); i++) {
            values[i * 3] = unmodifiable.get(unmodifiableIndices.get(i) * 3);
            values[i * 3 + 1] = unmodifiable.get(unmodifiableIndices.get(i) * 3 + 1);
            values[i * 3 + 2] = unmodifiable.get(unmodifiableIndices.get(i) * 3 + 2);
        }
        
        return values;
    }
    
    /**
     * Get a copy of the normals in an array.
     * @return 
     */
    public final float[] getNormals() {
        /* Create an unmodifiable copy of the list to make it thread safe */
        final List<Float> unmodifiable = Collections.unmodifiableList(normals);
        final List<Integer> unmodifiableIndices = Collections.unmodifiableList(normalIndices);
        final float[] values = new float[unmodifiableIndices.size() * 3];

        for (int i = 0; i < unmodifiableIndices.size(); i++) {
            values[i * 3] = unmodifiable.get(unmodifiableIndices.get(i) * 3);
            values[i * 3 + 1] = unmodifiable.get(unmodifiableIndices.get(i) * 3 + 1);
            values[i * 3 + 2] = unmodifiable.get(unmodifiableIndices.get(i) * 3 + 2);
        }
        
        return values;
    }
    
    /**
     * Get a copy of the texcoords in an array.
     * @return 
     */
    public final float[] getTexcoords() {
        /* Create an unmodifiable copy of the list to make it thread safe */
        final List<Float> unmodifiable = Collections.unmodifiableList(texcoords);
        final List<Integer> unmodifiableIndices = Collections.unmodifiableList(texcoordIndices);
        final float[] values = new float[unmodifiableIndices.size() * 2];
        
        for (int i = 0; i < unmodifiableIndices.size(); i++) {
            values[i * 2] = unmodifiable.get(unmodifiableIndices.get(i) * 2);
            values[i * 2 + 1] = unmodifiable.get(unmodifiableIndices.get(i) * 2 + 1);
        }
                
        return values;
    }
    
    /**
     * Get a copy of the vertex indices in an array.
     * @return 
     */
    public final int[] getIndices() {
        /* Create an unmodifiable copy of the list to make it thread safe */
        final List<Integer> unmodifiable = Collections.unmodifiableList(vertexIndices);
        final int[] values = new int[unmodifiable.size()];
        
        for (int i = 0; i < values.length; i++) {
            values[i] = unmodifiable.get(i);
        }
        
        return values;
    }

}
