/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.lidar.amapvox.engine3d.resources.postprocesses.base;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL3bc;
import java.io.IOException;
import org.joml.Vector2i;
import fr.amap.lidar.amapvox.engine3d.resources.Resource;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.Geometry;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.engine3d.resources.shaders.Shader;
import net.jcip.annotations.ThreadSafe;

/**
 * Post processing class to add to a camera.
 */
@ThreadSafe
public abstract class PostProcess extends Resource {

    private final Mesh mesh;
    
    private final Shader shader;
    private int inputTexture;
    private FrameBuffer frameBuffer;
    private boolean finalProcess;
    
    private final Vector2i resolution;
    private int resolutionLocation;

    /**
     * Create an empty post process object with a size of 500/500px.
     * @param identifier
     * @param hdr
     * @throws IOException 
     */
    public PostProcess(final String identifier, final boolean hdr) throws IOException {
        super(identifier);
        this.frameBuffer = new FrameBuffer("PostProcessFrameBuffer", 500, 500);
        this.finalProcess = false;
        this.shader = new Shader(identifier + "Shader");
        this.shader.setVertexSources("/shaders/postprocessing/generic/vertex.glsl");
        mesh = new Mesh("QuadMesh");
        mesh.setGeometry(Geometry.loadFromOBJ("/meshes/quad.obj"));
        this.resolution = new Vector2i();
    }
    
    /**
     * Load a GLSL fragment shader file and set it as the new fragment sources.
     * @param path 
     */
    protected final void setFragmentSources(final String path) {
        shader.setFragmentSources("/shaders/postprocessing/include/fragment.glsl", path);
    }
    
    /**
     * Resize the framebuffer.
     * @param gl
     * @param width
     * @param height 
     */
    public final void setSize(final GL3bc gl, final int width, final int height) {
        if (frameBuffer != null) {
            frameBuffer.unload(gl);
        }
        
        frameBuffer = new FrameBuffer("PostProcessFrameBuffer", width, height);
        frameBuffer.load(gl);
        
        synchronized(this) {
            this.resolution.x = width;
            this.resolution.y = height;
        }
    }

    public final synchronized void setFinalProcess(final boolean finalProcess) {
        this.finalProcess = finalProcess;
    }

    public final synchronized void setInputTexture(final int texture) {
        this.inputTexture = texture;
    }
    
    public final synchronized int getOutputTexture() {
        if (frameBuffer == null) {
            throw(new NullPointerException("Frame buffer is null"));
        }
        
        return frameBuffer.getColorTexture();
    }

    @Override
    protected void onLoad(final GL3bc gl) {
        shader.load(gl);
        mesh.load(gl);
        
        resolutionLocation = gl.glGetUniformLocation(shader.getProgram(), "resolution");
    }

    @Override
    protected void onBind(final GL3bc gl) {
        /* Set the default framebuffer if this one is null or this process 
        if the final process */
        if (frameBuffer != null && !finalProcess) {
            frameBuffer.bind(gl);
            
            /* Clear the color of the framebuffer by a transparent one */
            gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
            gl.glClear(GL3bc.GL_COLOR_BUFFER_BIT);
        } else {
            /* Bind the default framebuffer for the final process */
            /* This effect will be added to the before 
            drawed finals post-processes */
            gl.glBindFramebuffer(GL3bc.GL_FRAMEBUFFER, 0);
        }
        
        /* Bind the shader and the mesh */
        shader.bind(gl);
        mesh.bind(gl);
        
        /* Send the screen resolution */
        gl.glUniform2f(resolutionLocation, resolution.x, resolution.y);
        
        /* Send the input texture */
        gl.glActiveTexture(GL3bc.GL_TEXTURE0);
        gl.glEnable(GL3bc.GL_TEXTURE_2D);
        gl.glBindTexture(GL3bc.GL_TEXTURE_2D, inputTexture);
    }
    
    public final synchronized void draw(final GL3bc gl) {        
        /* Rescale to draw in the entire screen */
        gl.glViewport(0, 0, resolution.x, resolution.y);

        /* Clear the depth buffer */
        gl.glClear(GL.GL_DEPTH_BUFFER_BIT);
           
        /* Disable the depth test */
        gl.glDisable(GL.GL_DEPTH_TEST);
        
        /* Disable blending */
        gl.glDisable(GL3bc.GL_BLEND);
        
        /* Draw the screen square mesh */
        mesh.draw(gl);
    }

    @Override
    protected void onUnbind(final GL3bc gl) {
        shader.unbind(gl);
        mesh.unbind(gl);
    }

    @Override
    protected void onUnload(final GL3bc gl) {
        shader.unload(gl);
        mesh.unload(gl);
    }

}
